using GroupLab.Networking;
using MEdiaSpaceCore;
using System;
using System.Drawing;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace MEdiaSpaceCanmore
{
    public class FormController
    {
        // Form instances (small, medium, large)
        private static SmallForm _smallForm;
        private static MediumForm _mediumForm;
        private static LargeForm _largeForm;
        private static Form _activeForm;

        // Private status variables
        private static bool _connected;
        private static bool _talking;

        // Webcam source variable
        private static VideoWrapper _videoWrapper;

        // Audio wrapper variable
        private static AudioWrapper _audioWrapper;

        // Shared dictionary variables
        private static SharedDictionary _sd;
        private static Subscription _networkPortSub;
        private static Subscription _calgaryUserHandleSub;

        // SharedPhidgets variables
        private static GroupLab.SharedPhidgets.ConnectionManager _connectionManager;
        private static System.ComponentModel.IContainer _components;
        private static GroupLab.SharedPhidgets.InterfaceKit _interfaceKit;


        internal static void Initialize(MediumForm mainForm)
        {
            _smallForm = new SmallForm();
            _mediumForm = mainForm;
            _largeForm = new LargeForm();

            _activeForm = mainForm;

            _connected = false;
            _talking = false;
            _sd = null;

            SetConnected(false);
            SetUIEnabled(false);
            setDoorEnabled(false);
            SetCallStatus(false);
            SetLocalImage(null);
            SetRemoteImage(null);

            Variables.maxFPS = Constants.DEFAULT_SLOW_FRAMERATE;
            _videoWrapper = new VideoWrapper();
            _videoWrapper.ImageSent += new VideoWrapper.LocalImageSentHandler(SetLocalImage);
            _videoWrapper.ImageReceived += new VideoWrapper.RemoteImageSentHandler(SetRemoteImage);
            _videoWrapper.Start();

            // 
            // SharedPhidgets setup
            // 
            _connectionManager = new GroupLab.SharedPhidgets.ConnectionManager();
            _components = new System.ComponentModel.Container();
            _interfaceKit = new GroupLab.SharedPhidgets.InterfaceKit(_components);
            _interfaceKit.AutoAttach = true;
            _interfaceKit.ComponentDescriptionPrefix = null;
            _interfaceKit.FilterSerialNumbers.Add("31246");
            _interfaceKit.GUID = "59aed5ca-3bc6-431e-97cb-690c4fe66327";
            _interfaceKit.Invisible = false;
            _interfaceKit.OverrideDesignTimeLock = false;
            _interfaceKit.Attach += new System.EventHandler(interfaceKit_Attach);
            _interfaceKit.Detach += new System.EventHandler(interfaceKit_Detach);
            _interfaceKit.SensorChangeAsynchronous += new GroupLab.SharedPhidgets.IndexSensorChangeEventHandler(interfaceKit_SensorChangeAsynchronous);
        }

        public static void Shutdown()
        {
            _smallForm.Hide();
            _smallForm.Dispose();

            _largeForm.Hide();
            _largeForm.Dispose();

            _mediumForm.Hide();
            _mediumForm.Dispose();
        }

        public static Form ActiveForm()
        {
            return _activeForm;
        }


        //
        // Source/Sink callbacks
        //
        static void audioSource_AudioSourceStatusChanged(bool talking)
        {
            _talking = talking;
            SetCallStatus(_talking);

            ClearChatSpool();
        }


        //
        // Shared Dictionary callbacks
        //
        static void sd_Opened(object sender, EventArgs e)
        {
            _networkPortSub = new Subscription();
            _networkPortSub.BeginInit();
            _networkPortSub.Pattern = Constants.SDKEY_CALGARY_NETWORK_PORT;
            _networkPortSub.Dictionary = _sd;
            _networkPortSub.EndInit();

            _calgaryUserHandleSub = new Subscription();
            _calgaryUserHandleSub.BeginInit();
            _calgaryUserHandleSub.Pattern = Constants.SDKEY_CALGARY_SKYPE_USERNAME;
            _calgaryUserHandleSub.Dictionary = _sd;
            _calgaryUserHandleSub.EndInit();

            _audioWrapper = new AudioWrapper(null);
            _sd[Constants.SDKEY_CANMORE_SKYPE_USERNAME] = _audioWrapper.CurrentHandle();
            _audioWrapper.AudioStatusChanged += new AudioWrapper.AudioStatusChangedHandler(audioSource_AudioSourceStatusChanged);
            _audioWrapper.CallAttaching += new AudioWrapper.CallAttachingHandler(SetCallDisabled);
            SetCallDisabled(false);

            //int calgaryPort = (int)_sd.GetEntry(Constants.SDKEY_CALGARY_NETWORK_PORT).Value;
            int calgaryPort = Constants.DEFAULT_NETWORK_PORT;
            TcpClient client = new TcpClient();
            try
            {
                client.Connect(IPAddress.Parse(Constants.CALGARY_IP), calgaryPort);
            }
            catch (Exception error)
            {
                Console.WriteLine(error.Message);
                return;
            }

            _connected = true;
            SetConnected(true);
            SetUIEnabled(true);
            SetCallStatus(false);
            SetRemoteImage(null);
            ClearChatSpool();
            
            _videoWrapper.SetClient(client);
        }

        static void sd_Closed(object sender, EventArgs e)
        {
            _connected = false;
            SetConnected(false);
            SetUIEnabled(false);
            SetCallDisabled(true);
            SetCallStatus(false);
            SetLocalImage(null);
            SetRemoteImage(null);
            ClearChatSpool();

            _videoWrapper.SetClient(null);
            _sd = null;
        }
        

        //
        // SharedPhidgets callbacks
        //
        static void interfaceKit_Attach(object sender, EventArgs e)
        {
            setDoorEnabled(true);
        }

        static void interfaceKit_Detach(object sender, EventArgs e)
        {
            setDoorEnabled(false);
        }

        static void interfaceKit_SensorChangeAsynchronous(object sender, GroupLab.SharedPhidgets.IndexSensorChangeEventArgs e)
        {

        }


        //
        // Interaction sinks
        //
        public static void OnConnectClicked()
        {
            if (!_connected)
            {
                string url = "tcp://" + Constants.CALGARY_IP + ":" + Constants.SHARED_DICTIONARY_NAME;
                if (UrlPrompt.Show(ref url) == DialogResult.OK)
                {
                    _sd = new SharedDictionary();
                    _sd.Url = url;
                    _sd.Opened += new EventHandler(sd_Opened);
                    _sd.Closed += new EventHandler(sd_Closed);
                    _sd.Open(false);
                }

                _connectionManager.SharedDictionaryURL = "tcp://" + Constants.CALGARY_IP + ":sp";
                _connectionManager.OpenConnection();
            }
            else if (_sd != null)
            {
                _sd.Close(false);

                _connectionManager.CloseConnection();
            }
        }

        public static void OnDoorClicked()
        {
            setDoorEnabled(false);
            _interfaceKit.Outputs[0].State = true;
            Thread.Sleep(Constants.CALGARY_DOOR_RELAY_DELAY * 1000);
            _interfaceKit.Outputs[0].State = false;
            setDoorEnabled(true);
        }

        public static void OnDocumentClicked()
        {
        }

        public static void OnPhoneClicked()
        {
            if (!_talking && _sd.Contains(Constants.SDKEY_CALGARY_SKYPE_USERNAME)) {
                _audioWrapper.SetSinkUsername((String)_sd[Constants.SDKEY_CALGARY_SKYPE_USERNAME]);
                Thread audioSourceThread = new Thread(new ThreadStart(_audioWrapper.StartCall));
                audioSourceThread.IsBackground = true;
                audioSourceThread.Start();
            }
            else
                _audioWrapper.StopCall();
        }

        public static void OnToolsClicked()
        {
        }

        public static void OnMessageEntered(String message)
        {
            // TEST CODE
            AddChatMessage("Message: ", message);
            // END TEST CODE
        }


        //
        // Content distribution functions
        //
        public static void SetConnected(bool connected)
        {
            _smallForm.SetConnected(connected);
            _mediumForm.SetConnected(connected);
            _largeForm.SetConnected(connected);
        }

        public static void setDoorEnabled(bool enabled)
        {
            _mediumForm.SetDoorEnabled(enabled);
            _largeForm.setDoorEnabled(enabled);
        }

        public static void SetUIEnabled(bool enabled)
        {
            _mediumForm.SetUIEnabled(enabled);
            _largeForm.SetUIEnabled(enabled);
        }

        public static void SetCallDisabled(bool disabled)
        {
            _mediumForm.SetCallDisabled(disabled);
            _largeForm.SetCallDisabled(disabled);
        }

        public static void SetCallStatus(bool talking)
        {
            _smallForm.SetCallStatus(talking);
            _mediumForm.SetCallStatus(talking);
            _largeForm.SetCallStatus(talking);

            if (talking)
                Variables.maxFPS = Constants.DEFAULT_FAST_FRAMERATE;
            else
                Variables.maxFPS = Constants.DEFAULT_SLOW_FRAMERATE;
        }

        public static void SetMotionDetection(int value, DateTime timestamp)
        {
            _smallForm.SetMotionDetection(value, timestamp);
            _mediumForm.SetMotionDetection(value, timestamp);
            _largeForm.SetMotionDetection(value, timestamp);
        }
        
        public static void SetRemoteImage(Image image)
        {
            // These are expensive, so just update the active window
            if (_activeForm == _smallForm)
                _smallForm.SetRemoteImage(image);
            else if (_activeForm == _mediumForm)
                _mediumForm.SetRemoteImage(image);
            else
                _largeForm.SetRemoteImage(image);
        }

        public static void SetLocalImage(Image image)
        {
            // These are expensive, so just update the active window
            if (_activeForm == _mediumForm)
                _mediumForm.SetLocalImage(image);
            else if (_activeForm == _largeForm)
                _largeForm.SetLocalImage(image);
        }

        public static void ClearChatSpool()
        {
            _mediumForm.ClearChatSpool();
            _largeForm.ClearChatSpool();
        }

        public static void AddChatMessage(String sender, String message)
        {
            _mediumForm.AddChatMessage(sender, message);
            _largeForm.AddChatMessage(sender, message);
        }


        //
        // Utility functions
        //
        public static void SwitchToSmall()
        {
            if (_activeForm == _smallForm)
                return;

            _activeForm.Hide();
            _smallForm.Show();

            _activeForm = _smallForm;
        }

        public static void SwitchToMedium()
        {
            if (_activeForm == _mediumForm)
                return;

            _activeForm.Hide();
            _mediumForm.Show();

            _activeForm = _mediumForm;
        }

        public static void SwitchToLarge()
        {
            if (_activeForm == _largeForm)
                return;

            _activeForm.Hide();
            _largeForm.Show();

            _activeForm = _largeForm;
        }
    }
}
