using EasyImages;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Text;

namespace MEdiaSpaceCore
{
    public class VideoWrapper
    {
        // Callback bits so the hosting UI can get a copy of images as they're sent out
        public delegate void LocalImageSentHandler(Image image);
        public event LocalImageSentHandler ImageSent;

        // Callback bits so the hosting UI can get a copy of images as they're sent out
        public delegate void RemoteImageSentHandler(Image image);
        public event RemoteImageSentHandler ImageReceived;

        // Callback bits so the hosting UI can be informed if the TcpClient is closed
        public delegate void ClientClosedHandler();
        public event ClientClosedHandler ClientClosed;

        // Networking bits (these are provided by the caller!)
        private TcpClient _client;
        private NetworkStream _stream;

        // Camera bits
        private EasyImages.CameraClient _camera;
        private long _previousFrameBroadcastTime;

        // The image buffer data we store from frame to frame
        private Bitmap _localImage;
        private int _imageBufferWidth, _imageBufferHeight, _imageBufferEncoding, _imageBufferDataLength;
        private byte[] _imageBufferArray;
        private bool _imageDataReady;

        // Stop-and-go protocol support
        private Object _mutex;
        private byte[] _tinyBuffer;

        public VideoWrapper()
        {
            // By default, we have no client, no stream and no received image watcher
            _client = null;
            _stream = null;
            ImageReceived = null;

            // Set up an object to use as a mutex for the image frame buffer
            _mutex = new Object();

            // And a buffer to receive the ACK packet
            _tinyBuffer = new byte[1];

            // Indicate that we're waiting for the camera (not the network)
            _imageDataReady = false;

            // And we've not yet sent a single frame
            _previousFrameBroadcastTime = 0;
        }

        public void Start()
        {
            _previousFrameBroadcastTime = 0;
            _camera = new EasyImages.CameraClient("DefaultCamera");
            _camera.ReceivedFrame += new CameraClient.ReceivedFrameEventHandler(camera_ReceivedFrame);
            _camera.FramesPerSecond = (int)Constants.DEFAULT_FAST_FRAMERATE;
            _camera.Start();
        }

        public void Stop()
        {
            // TODO: Implement
        }

        public void SetClient(TcpClient client)
        {
            _client = client;

            if (_client == null)
                _stream = null;
            else
                _stream = _client.GetStream();
        }

        private void camera_ReceivedFrame(object sender, CameraEventArgs e)
        {
            // If it's not been long enough since we sent the last frame, just stall for a capture cycle
            double timeSinceLastBroadcast = (System.DateTime.Now.Ticks - _previousFrameBroadcastTime) / 10000000.0;
            if (timeSinceLastBroadcast > 0.0 && timeSinceLastBroadcast < (1.0 / Variables.maxFPS))
                return;

            // If we haven't got anywhere for this frame to go, we also stall
            if (_stream == null)
                return;

            BitmapData data = e.Bitmap.LockBits(new Rectangle(0, 0, e.Bitmap.Width, e.Bitmap.Height),
                                                ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);

            // Store the image data to the buffer (we throw away frames we don't need)
            lock (_mutex)
            {
                // Store the image header field data
                _imageBufferWidth = e.Bitmap.Width;
                _imageBufferHeight = e.Bitmap.Height;
                _imageBufferEncoding = (int)Variables.currentProtocol;

                // Encode and store the image into the buffer
                switch (Variables.currentProtocol)
                {
                    case Constants.VIDEO_PROTOCOLS.RAW:
                        _imageBufferDataLength = e.Bitmap.Width * e.Bitmap.Height * Constants.BYTES_PER_PIXEL;

                        unsafe
                        {
                            _imageBufferArray = new byte[_imageBufferDataLength];
                            Marshal.Copy(data.Scan0, _imageBufferArray, 0, _imageBufferDataLength);
                        }
                        break;

                    case Constants.VIDEO_PROTOCOLS.JPEG:
                        EasyImages.JPEG encoded = new EasyImages.JPEG(e.Bitmap, Variables.jpegCompression);
                        _imageBufferArray = encoded.GetByteArray();

                        _imageBufferDataLength = _imageBufferArray.Length;
                        break;

                    default:
                        _imageBufferDataLength = 0;
                        break;
                }
            }
            e.Bitmap.UnlockBits(data);

            // Cache the image so that we can display it locally when it goes out on the wire
            _localImage = (Bitmap)e.Bitmap.Clone();

            // If the roadblock was because of the camera (as opposed to the network), send the packet ASAP
            if (!_imageDataReady && _stream != null)
            {
                _imageDataReady = true;
                stream_ReceivedBytes(null);
            }
        }

        private void stream_ReceivedBytes(IAsyncResult asyncResult)
        {
            byte[] dataBuffer;

            // Figure out what kind of packet we're looking at...
            if (asyncResult == null || _tinyBuffer[0] == (byte)Constants.NETWORK_PACKETS.ACK) {

                // If the buffer hasn't been repopulated yet, then just wait until we're ready to go
                if (_imageBufferDataLength == 0)
                {
                    _imageDataReady = false;
                    return;
                }

                // Echo our local image to the UI (this should happen whether or not we're all set
                // up to send it out someplace, so long as a callback is set up...)
                if (ImageSent != null)
                    ImageSent(_localImage);

                lock (_mutex)
                {
                    // Frame format:
                    // - Image packet header (1 byte)
                    // - Image width (4 bytes)
                    // - Image height (4 bytes)
                    // - Image encoding (4 bytes)
                    // - Image data length (4 bytes)
                    // - Image data (variable length)

                    // Send the image packet header
                    try
                    {
                        _stream.WriteByte((byte)Constants.NETWORK_PACKETS.VIDEO);
                    }
                    catch (IOException)
                    {
                        // If we can't write, the client must've disappeared out from under us
                        _client = null;
                        _stream = null;
                        _imageBufferDataLength = 0;
                        _imageDataReady = false;

                        if (ClientClosed != null)
                            ClientClosed();

                        return;
                    }

                    // Send the image width
                    dataBuffer = new byte[Constants.HEADER_FIELD_LENGTH];
                    dataBuffer = System.BitConverter.GetBytes(_imageBufferWidth);
                    _stream.Write(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);

                    // Send the image height
                    dataBuffer = System.BitConverter.GetBytes(_imageBufferHeight);
                    _stream.Write(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);

                    // Send the image encoding
                    dataBuffer = System.BitConverter.GetBytes(_imageBufferEncoding);
                    _stream.Write(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);

                    // Send the image data length
                    dataBuffer = System.BitConverter.GetBytes(_imageBufferDataLength);
                    _stream.Write(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);

                    // Send the image data
                    _stream.Write(_imageBufferArray, 0, _imageBufferArray.Length);

                    // Clear the buffer (so we don't send the same data repeatedly)
                    _imageBufferDataLength = 0;
                }

                // Record the time we sent the frame out
                _previousFrameBroadcastTime = System.DateTime.Now.Ticks;
            }
            else if (_tinyBuffer[0] == (byte)Constants.NETWORK_PACKETS.VIDEO) {
                // Capture our timing data
                // Frame format:
                // - Image width (4 bytes)
                // - Image height (4 bytes)
                // - Image encoding (4 bytes)
                // - Image data length (4 bytes)
                // - Image data (variable length)

                // Read the image width
                dataBuffer = new byte[Constants.HEADER_FIELD_LENGTH];
                _stream.Read(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);
                int imageWidth = System.BitConverter.ToInt32(dataBuffer, 0);

                // Read the image height
                _stream.Read(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);
                int imageHeight = System.BitConverter.ToInt32(dataBuffer, 0);

                // Read the image encoding
                _stream.Read(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);
                Constants.VIDEO_PROTOCOLS imageEncoding =
                    (Constants.VIDEO_PROTOCOLS)System.BitConverter.ToInt32(dataBuffer, 0);

                // Read the image data length
                _stream.Read(dataBuffer, 0, Constants.HEADER_FIELD_LENGTH);
                int imageSize = System.BitConverter.ToInt32(dataBuffer, 0);

                // Create an image buffer that's the right size to receive the image object
                byte[] imageBuffer = new byte[imageSize];
                int imageBytesRead = 0;

                // Read the image data
                while (imageBytesRead < imageSize)
                {
                    // Read the next chunk of the image into the image buffer
                    imageBytesRead += _stream.Read(imageBuffer, imageBytesRead, imageSize - imageBytesRead);
                }

                // Convert the image buffer back into a Bitmap object (different process for each encoding)
                Bitmap bmap = null;
                switch (imageEncoding)
                {
                    case Constants.VIDEO_PROTOCOLS.RAW:
                        unsafe
                        {
                            fixed (byte* imageBufferAddr = imageBuffer)
                            {
                                IntPtr imageBufferPtr = (IntPtr)imageBufferAddr;
                                bmap = new Bitmap(imageWidth, imageHeight, imageWidth * Constants.BYTES_PER_PIXEL,
                                                  System.Drawing.Imaging.PixelFormat.Format24bppRgb, imageBufferPtr);
                            }
                        }
                        break;

                    case Constants.VIDEO_PROTOCOLS.JPEG:
                        MemoryStream memStream = new MemoryStream();
                        memStream.Write(imageBuffer, 0, imageSize);
                        bmap = new Bitmap(memStream);
                        memStream.Close();
                        memStream.Dispose();
                        break;

                    default:
                        break;
                }

                // If it worked out and we have somewhere to send it, then send it
                if (bmap != null && ImageReceived != null)
                    ImageReceived(bmap);

                // In any case, tell the sender that we're ready for the next frame!
                _stream.WriteByte((byte)Constants.NETWORK_PACKETS.ACK);
            }
                    
            // And go back to sleep until we receive another packet
            if (_stream != null)
                _stream.BeginRead(_tinyBuffer, 0, 1, new AsyncCallback(stream_ReceivedBytes), null);
        }
    }
}
